package djmil.cordacheckers.contracts;

import djmil.cordacheckers.states.GameBoardState;
import djmil.cordacheckers.states.GameProposalState;
import djmil.cordacheckers.states.GameResultState;
import djmil.cordacheckers.states.GameState;
import net.corda.v5.base.annotations.ConstructorForDeserialization;
import net.corda.v5.base.annotations.CordaSerializable;
import net.corda.v5.base.types.MemberX500Name;
import net.corda.v5.ledger.utxo.Command;
import net.corda.v5.ledger.utxo.StateAndRef;
import net.corda.v5.ledger.utxo.transaction.UtxoLedgerTransaction;

import java.lang.reflect.Array;
import java.util.LinkedList;
import java.util.List;

import static djmil.cordacheckers.contracts.UtxoLedgerTransactionUtil.getSingleInputState;
import static djmil.cordacheckers.contracts.UtxoLedgerTransactionUtil.getSingleOutputState;

public class GameCommand implements Command {
    @CordaSerializable
    public static enum Action {
        GAME_PROPOSAL_CREATE,
        GAME_PROPOSAL_CANCEL,
        GAME_PROPOSAL_ACCEPT,
        GAME_PROPOSAL_REJECT,

        GAME_BOARD_MOVE,
        GAME_BOARD_SURRENDER,

        GAME_RESULT_CREATE;
    }

    public final Action action;
    public final List<Integer> move; // [0] from, [1] to

    public static class ActionException extends RuntimeException {
        public ActionException() {
            super("Bad GameCommand.Action value");
        }
    }

    // Serialisation service requires a default constructor
    public GameCommand() {
        this.action = null;
        this.move = null;
    }

    public GameCommand(Action action) {
        if (action == Action.GAME_BOARD_MOVE)
            throw new ActionException();

        this.action = action;
        this.move = null;
    }

    public GameCommand(List<Integer> move) {
        this.action = Action.GAME_BOARD_MOVE;
        this.move = move;
    }

    @ConstructorForDeserialization
    public GameCommand(Action action, List<Integer> move) {
        this.action = action;
        this.move = move;
    }

    public Action getAction() {
        return action;
    }

    public List<Integer> getMove() {
        return move;
    }
    

    /*
     * Session initiator/respondent
     */

    public MemberX500Name getObserver(StateAndRef<GameState> gameStateSar) {
        final GameState gameState = gameStateSar.getState().getContractState();
        return gameState.getCounterpartyName(getActor(gameState));
    }

    public MemberX500Name getActor(StateAndRef<GameState> gameStateSar) {
        return getActor(gameStateSar.getState().getContractState());
    }

    public MemberX500Name getObserver(GameState gameState) {
        return gameState.getCounterpartyName(getActor(gameState));
    }

    public MemberX500Name getActor(GameState gameState) {
        switch (action) {
            case GAME_PROPOSAL_CREATE:
            case GAME_PROPOSAL_CANCEL:
                if (gameState instanceof GameProposalState)
                    return ((GameProposalState)gameState).getIssuer();
                break;

            case GAME_PROPOSAL_REJECT:
                if (gameState instanceof GameProposalState)
                    return ((GameProposalState)gameState).getAcquier();
                break;
            
            case GAME_PROPOSAL_ACCEPT:
                if (gameState instanceof GameProposalState) // <<-- Session validation perspective
                    return ((GameProposalState)gameState).getAcquier();
                if (gameState instanceof GameBoardState)    // <<-- GameViewBuilder perspective
                    return ((GameBoardState)gameState).getMovePlayerName();
                break;

            case GAME_BOARD_MOVE:
                if (gameState instanceof GameBoardState)
                    return ((GameBoardState)gameState).getMovePlayerName();
                break;

            case GAME_BOARD_SURRENDER:
                if (gameState instanceof GameResultState)
                    return ((GameResultState)gameState).getOpponentName();
                break;
            
            case GAME_RESULT_CREATE:
                if (gameState instanceof GameResultState)
                    return ((GameResultState)gameState).getWinnerName();
                break;
            
            default:
                throw new ActionException();
        }

        throw new RuntimeException(action +": unexpected GameState type " +gameState.getClass());
    }

    /*
     * Transaction validation
     */

    public static void validateGameProposalCreate(UtxoLedgerTransaction trx) {
        requireThat(trx.getInputContractStates().isEmpty(), CREATE_INPUT_STATE);
        requireThat(trx.getOutputContractStates().size() == 1, CREATE_OUTPUT_STATE);

        GameProposalState outputState = getSingleOutputState(trx, GameProposalState.class);

        requireThat(outputState.getAcquierColor() != null, NON_NULL_RECIPIENT_COLOR);
    }

    public static void validateGameProposalAccept(UtxoLedgerTransaction trx) {
        requireThat(trx.getInputContractStates().size() == 1, ACCEPT_INPUT_STATE);
        requireThat(trx.getOutputContractStates().size() == 1, ACCEPT_OUTPUT_STATE);

        GameProposalState inGameProposal = getSingleInputState(trx, GameProposalState.class);
        GameBoardState outGameBoard      = getSingleOutputState(trx, GameBoardState.class);

        requireThat(outGameBoard.getParticipants().containsAll(inGameProposal.getParticipants()), IN_OUT_PARTICIPANTS);
    }

    public static void validateGameProposalReject(UtxoLedgerTransaction trx) {
        requireThat(trx.getInputContractStates().size() == 1, REJECT_INPUT_STATE);
        requireThat(trx.getOutputContractStates().isEmpty(), REJECT_OUTPUT_STATE);

        getSingleInputState(trx, GameProposalState.class);
    }

    public static void validateGameProposalCancel(UtxoLedgerTransaction trx) {
        requireThat(trx.getInputContractStates().size() == 1, CANCEL_INPUT_STATE);
        requireThat(trx.getOutputContractStates().isEmpty(), CANCEL_OUTPUT_STATE);
        
        getSingleInputState(trx, GameProposalState.class);
    }

    public static void validateGameBoardSurrender(UtxoLedgerTransaction trx) {
        requireThat(trx.getInputContractStates().size() == 1, SURRENDER_INPUT_STATE);
        final var inGameBoardState = getSingleInputState(trx, GameBoardState.class);

        requireThat(trx.getOutputContractStates().size() == 1, SURRENDER_OUTPUT_STATE);
        final var outGameResultState = getSingleOutputState(trx, GameResultState.class);


        List<MemberX500Name> inActors = new LinkedList<MemberX500Name>(List.of(
            inGameBoardState.getWhitePlayerName(),
            inGameBoardState.getBlackPlayerName())
        );

        List<MemberX500Name> outActors = new LinkedList<MemberX500Name>(List.of(
            outGameResultState.getWinnerName(),
            outGameResultState.getOpponentName())
        );

        requireThat(inActors.containsAll(outActors) && outActors.containsAll(inActors), IN_OUT_PARTICIPANTS);

        final var activePlayerName = outGameResultState.getCounterpartyName(outGameResultState.getWinnerName());
        requireThat(inGameBoardState.getMovePlayerName().compareTo(activePlayerName) == 0, SURRENDER_ACTOR + "expected " +inGameBoardState.getMovePlayerName() +" actual " +activePlayerName);
    }

    public static void validateGameBoardMove(UtxoLedgerTransaction trx) {
        requireThat(trx.getInputContractStates().size() == 1, MOVE_INPUT_STATE);
        final var inGameBoardState = getSingleInputState(trx, GameBoardState.class);

        requireThat(trx.getOutputContractStates().size() == 1, MOVE_OUTPUT_STATE);
        final var outGameBoardState = getSingleOutputState(trx, GameBoardState.class);

        requireThat(inGameBoardState.getWhitePlayerName().compareTo(outGameBoardState.getWhitePlayerName()) == 0, IN_OUT_PARTICIPANTS);
        requireThat(inGameBoardState.getBlackPlayerName().compareTo(outGameBoardState.getBlackPlayerName()) == 0, IN_OUT_PARTICIPANTS);
    }

    public static void validateGameResultCreate(UtxoLedgerTransaction trx) {
        throw new RuntimeException("Unimplemented");
    }

    public static void requireThat(boolean asserted, String errorMessage) {
        if (!asserted) {
            throw new IllegalStateException("Failed requirement: " + errorMessage);
        }
    }

    static final String REQUIRE_SINGLE_COMMAND = "Require a single command";
    static final String IN_OUT_PARTICIPANTS = "Input participants should represent output participants";

    static final String CREATE_INPUT_STATE  = "Create command should have no input states";
    static final String CREATE_OUTPUT_STATE = "Create command should output exactly one GameProposal state";
    static final String NON_NULL_RECIPIENT_COLOR = "GameProposal.recipientColor field can not be null";

    static final String REJECT_INPUT_STATE  = "Reject command should have exactly one GameProposal input state";
    static final String REJECT_OUTPUT_STATE = "Reject command should have no output states";

    static final String CANCEL_INPUT_STATE  = "Cancel command should have exactly one GameProposal input state";
    static final String CANCEL_OUTPUT_STATE = "Cancel command should have no output states";

    static final String ACCEPT_INPUT_STATE  = "Accept command should have exactly one GameProposal input state";
    static final String ACCEPT_OUTPUT_STATE = "Accept command should have exactly one GameBoard output state";

    static final String SURRENDER_INPUT_STATE = "SURRENDER command should have exactly one GameBoardState input state";
    static final String SURRENDER_OUTPUT_STATE = "SURRENDER command should have exactly one GameResultState output state";
    static final String SURRENDER_ACTOR = "Only active player can SURRENDER";

    static final String MOVE_INPUT_STATE = "MOVE command should have exactly one GameBoardState input state";
    static final String MOVE_OUTPUT_STATE = "MOVE command should have exactly one GameBoardState output state";
    static final String MOVE_OUT_BOARD = "MOVE command: move checkers rules violation";
    static final String MOVE_OUT_COLOR = "MOVE command: moveColor checkers rules violation";
}
