package djmil.cordacheckers.states;

import java.security.PublicKey;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import djmil.cordacheckers.contracts.GameBoardContract;
import djmil.cordacheckers.states.Piece.Color;
import net.corda.v5.base.annotations.ConstructorForDeserialization;
import net.corda.v5.base.types.MemberX500Name;
import net.corda.v5.ledger.utxo.BelongsToContract;

@BelongsToContract(GameBoardContract.class)
public class GameBoardState extends GameState {
    
    private final MemberX500Name whitePlayerName;
    private final MemberX500Name blackPlayerName;
    
    private final Piece.Color moveColor;
    private final Integer moveNumber;
    private final Map<Integer, Piece> board;

    public GameBoardState(GameProposalState gameProposalState) {
        super(gameProposalState.gameUuid, gameProposalState.message, gameProposalState.participants);

        this.whitePlayerName = gameProposalState.getWhitePlayerName();
        this.blackPlayerName = gameProposalState.getBlackPlayerName();
        
        // Initial GameBoard state
        this.moveColor = Piece.Color.WHITE;
        this.moveNumber = 0;
        this.board = new LinkedHashMap<Integer, Piece>(initialBoard);
    }

    public GameBoardState(
        GameBoardState oldGameBoardState, Map<Integer, Piece> newBoard, Piece.Color moveColor) {
        super(oldGameBoardState.gameUuid, oldGameBoardState.message, oldGameBoardState.participants);

        this.whitePlayerName = oldGameBoardState.getWhitePlayerName();
        this.blackPlayerName = oldGameBoardState.getBlackPlayerName();
        
        // Initial GameBoard state
        this.moveColor = moveColor;
        this.moveNumber = oldGameBoardState.getMoveNumber() +1;
        this.board = newBoard;
    }
 
    @ConstructorForDeserialization
    public GameBoardState(MemberX500Name whitePlayerName, MemberX500Name blackPlayerName, 
            Color moveColor, Integer moveNumber, Map<Integer, Piece> board, String message, 
            UUID gameUuid, List<PublicKey> participants) {
        super(gameUuid, message, participants);

        this.whitePlayerName = whitePlayerName;
        this.blackPlayerName = blackPlayerName;
        this.moveColor = moveColor;
        this.moveNumber = moveNumber;
        this.board = board;
    }

    public MemberX500Name getMovePlayerName() {
        switch (moveColor) {
            case WHITE:
                return whitePlayerName;
            case BLACK:
                return blackPlayerName;
            default:
                throw new Piece.Color.UnknownException();
        }
    }

    public Piece.Color getMoveColor() {
        return moveColor;
    }

    public Integer getMoveNumber() {
        return moveNumber;
    }

    public Map<Integer, Piece> getBoard() {
        return board;
    }

    @Override
    public MemberX500Name getWhitePlayerName() {
        return whitePlayerName;
    }

    @Override
    public MemberX500Name getCounterpartyName(MemberX500Name myName) throws NotInvolved {
        if (whitePlayerName.compareTo(myName) == 0)
            return blackPlayerName;
    
        if (blackPlayerName.compareTo(myName) == 0)
            return whitePlayerName;
        
        throw new GameState.NotInvolved(myName, GameBoardState.class, this.gameUuid);
    }

    // TODO: move to contract
    public final static Map<Integer, Piece> initialBoard =  Map.ofEntries(
        // Inspired by Checkers notation rules: https://www.bobnewell.net/nucleus/checkers.php
        Map.entry( 1, new Piece(Piece.Color.BLACK, Piece.Type.MAN)),
        Map.entry( 2, new Piece(Piece.Color.BLACK, Piece.Type.MAN)),
        Map.entry( 3, new Piece(Piece.Color.BLACK, Piece.Type.MAN)),
        Map.entry( 4, new Piece(Piece.Color.BLACK, Piece.Type.MAN)),
        Map.entry( 5, new Piece(Piece.Color.BLACK, Piece.Type.MAN)),
        Map.entry( 6, new Piece(Piece.Color.BLACK, Piece.Type.MAN)),
        Map.entry( 7, new Piece(Piece.Color.BLACK, Piece.Type.MAN)),
        Map.entry( 8, new Piece(Piece.Color.BLACK, Piece.Type.MAN)),
        Map.entry( 9, new Piece(Piece.Color.BLACK, Piece.Type.MAN)),
        Map.entry(10, new Piece(Piece.Color.BLACK, Piece.Type.MAN)),
        Map.entry(11, new Piece(Piece.Color.BLACK, Piece.Type.MAN)),
        Map.entry(12, new Piece(Piece.Color.BLACK, Piece.Type.MAN)),

        Map.entry(21, new Piece(Piece.Color.WHITE, Piece.Type.MAN)),
        Map.entry(22, new Piece(Piece.Color.WHITE, Piece.Type.MAN)),
        Map.entry(23, new Piece(Piece.Color.WHITE, Piece.Type.MAN)),
        Map.entry(24, new Piece(Piece.Color.WHITE, Piece.Type.MAN)),
        Map.entry(25, new Piece(Piece.Color.WHITE, Piece.Type.MAN)),
        Map.entry(26, new Piece(Piece.Color.WHITE, Piece.Type.MAN)),
        Map.entry(27, new Piece(Piece.Color.WHITE, Piece.Type.MAN)),
        Map.entry(28, new Piece(Piece.Color.WHITE, Piece.Type.MAN)),
        Map.entry(29, new Piece(Piece.Color.WHITE, Piece.Type.MAN)),
        Map.entry(30, new Piece(Piece.Color.WHITE, Piece.Type.MAN)),
        Map.entry(31, new Piece(Piece.Color.WHITE, Piece.Type.MAN)),
        Map.entry(32, new Piece(Piece.Color.WHITE, Piece.Type.MAN))
        );
}
